"""
Module containing estimators for object detection.
"""
from art.estimators.object_detection.object_detector import ObjectDetectorMixin

from art.estimators.object_detection.PyTorchFasterRCNN import PyTorchFasterRCNN
